/* include gnu.gpl, (*) 2000 Micha 'malekith' Moskal <malekith@topnet.pl>. 
 * $Id: curses.c,v 1.1.1.1 2000/07/29 19:53:07 malekith Exp $ */

#define DOUBLE_ESC
#define _GNU_SOURCE

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <sys/ioctl.h>

#include <ncurses/ncurses.h>

#include "term.h"

int s_width, s_height;
int scroll_top, scroll_bottom;

static int color;
static int dirty;
static int cur_x, cur_y;
static int stick_x = -1, stick_y;

static int attr_map[256];

void set_color(int vga)
{
	color = vga;
	attrset(attr_map[vga&0xff]);
}

void tputchar2(int ch, int col)
{
	//if (cur_x == s_width) { cur_x = 0; cur_y++; }
	if (cur_x++ < 0 || cur_x > s_width || cur_y < 0 || cur_y >= s_height)
		return;
	if (col != -1 && col != color)
		set_color(col);
	addch(ch);
	dirty = 1;
}

void tputchar(int ch)
{
	tputchar2(ch, -1);
}

int tprintf(const char *str, ... )
{
	char *b, *p;
	int r;
	va_list ap;
	
	va_start(ap, str);
	r = vasprintf(&b, str, ap);
	if (!b) {
		addstr("vasprintf: out of memory !");
		dirty = 1;
		return -1;
	}
	va_end(ap);
	
	for (p = b; *p; p++)
		tputchar(((int)*p) & 0xff);
	free(b);
	return r;
}

static void resize_handler(int x)
{
	struct winsize size;
	
	(void)x;
	
	s_height = LINES;
	s_width = COLS;
	
	if (ioctl(1, TIOCGWINSZ, &size) == 0) {
		resizeterm(size.ws_row, size.ws_col);
		s_height = LINES;
		s_width = COLS;
		f_repaint();
	}
	
	scroll_top = 0;
	scroll_bottom = s_height - 4;
	
	signal(SIGWINCH, resize_handler);
}

void f_repaint()
{
	wrefresh(curscr);
	dirty = 0;
}

void repaint()
{
	if (dirty) {
		if (stick_x != -1)
			go(stick_x, stick_y);
		refresh();
	}
	dirty = 0;
}

void go(int x, int y)
{
	move(y < 0 ? 0 : (y > s_height ? s_height : y), 
	     x < 0 ? 0 : (x > s_width ? s_width : x));
	move(y, x);
	cur_x = x;
	cur_y = y;
	dirty = 1;
}

void stick_cursor()
{
	stick_x = cur_x;
	stick_y = cur_y;
}

/* schamelessly stolen from linux tvision port */
static void init_color_colors()
{
	int vgac[] = {
		COLOR_BLACK, COLOR_BLUE, COLOR_GREEN, COLOR_CYAN,
		COLOR_RED, COLOR_MAGENTA, COLOR_YELLOW, COLOR_WHITE,
	};
	int back, fore, bold, i;
	
	start_color();
	
	i = 0;
	for (back = 0; back < 8; back++)
		for (fore = 7; fore >= 0; fore--)
			init_pair(i++, vgac[fore], vgac[back]);

	for (i = 0; i < 256; i++) {
		back = (i >> 4) & 0x07;
		bold = i & 0x08;
		fore = i & 0x07;
		attr_map[i] = COLOR_PAIR((7 - fore) + back * 8);
		if (bold)
			attr_map[i] |= A_BOLD;
	}
}

static void init_mono_colors()
{
	int i;
	/* subjective brightness of vga colors to decide whatever */
	/* to use A_REVERSE and/or A_BOLD */
	int vgab[] = {
		/*BLACK*/ 0, /*BLUE*/ 4, /*GREEN*/ 5, /*CYAN*/ 8, /*RED*/ 3, 
		/*MAGENTA*/ 7, /*BROWN*/ 2, /* LIGHTGRAY*/ 11, /*DARKGRAY*/ 1, 
		/*LIGHTBLUE*/ 6, /*LIGHTGREEN*/ 12, /*LIGHTCYAN*/ 13, 
		/*LIGHTRED*/ 9, /*LIGHTMAGENTA*/ 10, /*YELLOW*/ 14, 
		/*WHITE*/ 15
	};
	
	for (i = 0; i < 256; i++) {
		attr_map[i] = vgab[(i>>4)&0xf] > vgab[i&0xf] ? A_REVERSE : 0;
		if ((i&0xf) > 11)
			attr_map[i] |= A_BOLD;
	}
}

static void xinit_color()
{
	if (has_colors())
		init_color_colors();
	else
		init_mono_colors();
}

void init_curses()
{
	int n = 25;
	while (n-- > 0) printf("\r\n");
	/* allow iso-8859-[12] chars - we don't want vga rom frames and some */
	/* ugly progs (e.g. BitchX ;) tends to left console in such state */
	if (getenv("TERM") && strcmp(getenv("TERM"),"linux") == 0)
		printf("\e(B"); 

#ifndef DOUBLE_ESC
	setenv("ESCDELAY", "100", 0);
#endif

	initscr();
	keypad(stdscr, TRUE);
	raw();
	noecho(); 
	timeout(0);
	refresh();
	meta(stdscr, TRUE);
	idlok(stdscr, TRUE);
	
	/* mousemask(REPORT_MOUSE_POSITION, 0); */
	
	xinit_color();
	resize_handler(0);
	dirty = 1;
}

void kill_curses()
{
	endwin();
	printf("\r");
}

int _keyb()
{
	int k;
#ifdef DOUBLE_ESC
	static int last;
#endif

	k = getch();

	if (k == ERR)
		return 0;
		
	if (k == KEY_SUSPEND)
		raise(SIGTSTP);
	
	switch (k) {
	case KEY_LEFT: k = key_left; break;
	case KEY_RIGHT: k = key_right; break;
	case KEY_UP: k = key_up; break;
	case KEY_DOWN: k = key_down; break;
	case KEY_HOME: k = key_home; break;
	case KEY_END: k = key_end; break;
	case KEY_IC: k = key_ins; break;
	case KEY_DC: k = key_del; break;
	case KEY_NPAGE: k = key_pgdn; break;
	case KEY_PPAGE: k = key_pgup; break;
	case KEY_B2: k = key_5; break;
	case KEY_MOUSE: k = key_mouse; break;
	case 127:
	case KEY_BACKSPACE: k = '\b'; break;
	default:
		if (k >= KEY_F(0) && k <= KEY_F(30))
			k = k - KEY_F(0) + key_f0;
		break;
	}
	
#ifdef DOUBLE_ESC
	if (k == 27) {
		if (last == 27) {
			last = 0;
			return 27;
		} else {
			last = 27;
			return 0;
		}
	} else {
		if (last == 27)
			k |= key_meta;
		last = k;
		return k;
	}
#else
	if (k == 27) {
		k = getch();
		if (k != ERR && k != 27)
			return k | key_meta;
		else
			return 27;
	} else
		return k;
#endif
}

/*
	(-1 -1) (0 -1) (1 -1)   (+-) (-) (-+)   0 10 20
	(-1  0) (0  0) (1  0)   (|-) (+) (-|)   1 11 21
	(-1  1) (0  1) (1  1)   (+-) (-) (-+)   2 12 22
	
	(2, *)  (-) 
	(3, *)  (|)
*/
int get_frame(int x, int y)
{
	switch (x) {
	case 2: return ACS_HLINE;
	case 3: return ACS_VLINE;
	}
	
	y += (x + 1) * 10 + 1;
	switch (y) {
#define a(s,b) case s: return ACS_ ## b ;
	a(0,ULCORNER);
	a(2,LLCORNER);
	a(20,URCORNER);
	a(22,LRCORNER);
	a(1,LTEE);
	a(10,TTEE);
	a(21,RTEE);
	a(12,BTEE);
	a(11,PLUS);
#undef a
	}
	return '%';
}

void scroll_up(int n)
{
	setscrreg(scroll_top, scroll_bottom);
	//go(0,0);insdelln(-n);
	scrollok(stdscr, TRUE);
	scrl(n);
	setscrreg(0, s_height - 1);
	scrollok(stdscr, FALSE);
}

void cls()
{
	int n;
	
	go(0,0);
	n = s_width * s_height;
	while (n-- > 0)
		addch(' ');
	go(0,0);
		
	dirty = 1;
}
