/* include gnu.gpl, (*) 2000 Micha 'malekith' Moskal <malekith@topnet.pl>. 
 * $Id: cproc.c,v 1.14 2000/08/26 01:31:12 malekith Exp $ */

#include "vm.h"
#include "proc.h"

#include <time.h>
#include <unistd.h>

hash *cprocs;

var *exec_cproc(struct proc *p, frame *f)
{
	return ((cproc)(p->cproc))(f);
}

struct proc *find_cproc(const char *name, u32 h)
{
	var *v;
	
	v = hash_find_ex2(cprocs, name, h);
	
	return (struct proc*)(v ? v->var_ptr : 0);
}

void add_cproc(const char *name, const char *proto, cproc exec)
{
	struct proc *p;
	
	p = (struct proc*)xmalloc(sizeof(struct proc));
	p->args = bc_make_arg_list(proto);
	p->body = 0;
	p->cproc = (void*)exec;
	p->name = xstrdup(name);
	
	hash_add(cprocs, new_pv(p->name, p));
}

char *getsarg(frame *f, const char *name)
{
	var *p;
	
	p = hash_find_ex(f->stack, name);
	
	if (p == 0)
		abort();
		
	if (p->what != V_STRING)
		abort();

	return p->var_string;
}

int getiarg(frame *f, const char *name)
{
	var *p;
	
	p = hash_find_ex(f->stack, name);
	
	if (p == 0)
		abort();
		
	if (p->what != V_INT)
		abort();

	return p->var_int;
}

static var *c_print(frame *f)
{
	printf("%s",getsarg(f,"s"));
	return 0;
}

static var *c_exists(frame *f)
{
	return new_iv(0, access(getsarg(f, "s"), F_OK) == 0);
}

var *source_file(const char *fn, frame *f);
var *veval(char *input, frame *f);

static var *c_eval(frame *f)
{
	return veval(getsarg(f, "s"), f);
}

static var *c_source(frame *f)
{
	return source_file(getsarg(f, "s"), f);
}

static var *c_undef(frame *f)
{
	(void)f;
	return 0;
}

static var *c_printd(frame *f)
{
	printf("%d",getiarg(f,"i"));
	return 0;
}

static var *c_atoi(frame *f)
{
	return new_iv(0, atoi(getsarg(f, "s")));
}

static var *c_rand(frame *f)
{
	(void)f;
		/* limit to positive numbers */
	return new_iv(0, arc4random() & 0x7fffffff);
}

static var *c_randstring(frame *f)
{
	char *l = getsarg(f, "l");
	int n = getiarg(f, "n");
	var *v;
	char *p;
	int len;
	
	v = new_svn(0, n);
	p = v->var_string;
	len = strlen(l);
	while (n--)
		*p++ = l[arc4random() % len];
	*p = 0;
	return v;
}

static var *c_itoa(frame *f)
{
	char *p;
	var *v;

	p = xf("%d", getiarg(f,"i"));
	v = new_sv(0, p);
	xfree(p);

	return v;
}

static var *c_time(frame *f)
{
	(void)f;
	return new_iv(0, time(0));
}

static var *c_ctime(frame *f)
{
	time_t t = getiarg(f,"t");
	
	return new_sv(0, ctime(&t));
}

static var *c_system(frame *f)
{
	char *s = getsarg(f,"s");

	return new_iv(0, system(s));
}

static var *c_backtick(frame *f)
{
	char *s = getsarg(f,"s");
	FILE *p;
	char buf[4096 + 1];
	char *x = 0;
	int n, got = 0;
	var *v;
	
	p = popen(s, "r");

	if (p == 0)
		return 0;
	for (;;) {
		n = fread(buf, 1, 4096, p);
		buf[n] = 0;
		if (n == 0)
			break;
		x = (char*)xrealloc(x, got + n + 1);
		memcpy(x + got, buf, n);
		got += n;
		x[got] = 0;
	}
	pclose(p);

	while (x[got-1] == '\n')
		x[--got] = 0;

	v = new_sv(0, "");
	xfree(v->var_string);
	v->var_string_size = got + 1;
	v->var_string = x;
	return v;
}

static var *c_getenv(frame *f)
{
	char *e;

	e = getenv(getsarg(f,"s"));
	if (e == 0)
		e = "";
	return new_sv(0, e);
}

static var *c_getenv_def(frame *f)
{
	char *e;

	e = getenv(getsarg(f,"s"));
	if (e == 0)
		e = getsarg(f, "def");
	return new_sv(0, e);
}

static var *c_tolower(frame *f)
{
	char *s = getsarg(f, "s");
	char *p;

	for (p = s; *p; p++)
		if (*p >= 'A' && *p <= 'Z')
			*p |= 0x20;
			
	return new_sv(0, s);
}

static var *c_toupper(frame *f)
{
	char *s = getsarg(f, "s");
	char *p;

	for (p = s; *p; p++)
		if (*p >= 'a' && *p <= 'z')
			*p &= ~0x20;
			
	return new_sv(0, s);
}

static var *c_strlen(frame *f)
{
	return new_iv(0, strlen(getsarg(f, "s")));
}

static var *c_str(frame *f)
{
	char s[2];

	s[0] = getiarg(f, "i");
	s[1] = 0;
	return new_sv(0, s);
}

var *c_qq(frame *f);

void init_cprocs()
{
	static int done = 0;
	
	if (done)
		return;
	done = 1;

	cprocs = hash_new(V_PTR);
	
	add_cproc("print", "string s", c_print);
	add_cproc("printd", "int i", c_printd);
	add_cproc("atoi", "string s", c_atoi);
	add_cproc("itoa", "int i", c_itoa);
	add_cproc("rand", "", c_rand);
	add_cproc("randstring", "string l, int n", c_randstring);
	add_cproc("time", "", c_time);
	add_cproc("ctime", "int t", c_ctime);
	add_cproc("system", "string s", c_system);
	add_cproc("bt", "string s", c_backtick);
	add_cproc("__back_tick", "string s", c_qq);
	add_cproc("undef", "", c_undef);
	add_cproc("getenv", "string s", c_getenv);
	add_cproc("getenv_def", "string s, string def", c_getenv_def);
	add_cproc("toupper", "string s", c_toupper);
	add_cproc("tolower", "string s", c_tolower);
	add_cproc("strlen", "string s", c_strlen);
	add_cproc("str", "int i", c_str);
	add_cproc("eval", "string s", c_eval);
	add_cproc("source", "string s", c_source);
	add_cproc("exists", "string s", c_exists);
}
